#if !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
#pragma once

#include <torch/csrc/Export.h>
#include <torch/csrc/jit/ir/ir.h>

namespace torch::jit {

TORCH_API const std::string& GetSerializedShapeFunctions();

TORCH_API const OperatorMap<std::string>& GetShapeFunctionMappings();

TORCH_API const OperatorMap<std::pair<std::string, std::string>>&
GetBoundedShapeMappings();

} // namespace torch::jit

#else
#error "This file should not be included when either TORCH_STABLE_ONLY or TORCH_TARGET_VERSION is defined."
#endif  // !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
